/************************************************************************
 * Copyright (c) 2008 by Gregor Ehrenstein (grx@pandacube.de)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ************************************************************************/

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;

namespace BoxEdit3D
{
    /// <summary>
    /// This is the main type for your game.
    /// </summary>
    public class BoxEdit3D : Microsoft.Xna.Framework.Game
    {
        // used to access the graphicsDevice
        GraphicsDeviceManager m_graphicsDeviceManager;

        // holds the information about all the boxes
        private Scene m_scene;

        // class used to display a translucent ground floor
        private Floor m_floor;

        // class that handles the logic behind user input
        private UserInput m_userInput;

        /// <summary>
        /// Constructor
        /// </summary>
        public BoxEdit3D()
        {
            // set preferred window size 
            m_graphicsDeviceManager = new GraphicsDeviceManager(this);
            m_graphicsDeviceManager.PreferredBackBufferWidth = 1280;
            m_graphicsDeviceManager.PreferredBackBufferHeight = 1024;

            // make sure window is not bigger than screen
            if (m_graphicsDeviceManager.PreferredBackBufferWidth > System.Windows.Forms.Screen.PrimaryScreen.WorkingArea.Width)  
                m_graphicsDeviceManager.PreferredBackBufferWidth = System.Windows.Forms.Screen.PrimaryScreen.WorkingArea.Width;
            if (m_graphicsDeviceManager.PreferredBackBufferHeight > System.Windows.Forms.Screen.PrimaryScreen.WorkingArea.Height)
                m_graphicsDeviceManager.PreferredBackBufferHeight = System.Windows.Forms.Screen.PrimaryScreen.WorkingArea.Height;

            // initialize variables
            m_scene = new Scene();
            m_floor = new Floor(m_graphicsDeviceManager);
            m_userInput = new UserInput(m_graphicsDeviceManager, m_scene, this.exitCallback);
            Content.RootDirectory = "Content";
        }

        /// <summary>
        /// Used as callback from the user input class to shut down this editor
        /// </summary>
        private void exitCallback()
        {
            this.Exit();
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            base.Initialize();
            this.IsMouseVisible = true;
            this.Window.AllowUserResizing = true;
            this.Window.ClientSizeChanged += new EventHandler(Window_ClientSizeChanged);
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // TODO: use seperate loader class that knows about all the classes
            // that have to initialized

            // load box model
            Box.loadResources(m_graphicsDeviceManager, Content);

            // load plane model
            m_floor.loadResources(Content);

            // load gizmo models
            TranslationGizmo.loadResources(Content);
            RotationGizmo.loadResources(Content);
            ScaleGizmo.loadResources(Content);
        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
            // free textures from texture manager
            Box.unloadResources();
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            m_userInput.Update(gameTime); 

            base.Update(gameTime);
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            m_graphicsDeviceManager.GraphicsDevice.Clear(Color.DarkSlateGray);

            // get viewProjection Matrix
            Matrix viewProjMatrix = m_userInput.getViewProjectionMatrix();

            // render scene
            m_scene.render(viewProjMatrix);

            // render gizmo
            m_userInput.renderGizmo(viewProjMatrix);

            // render plane
            m_floor.render(viewProjMatrix);

            base.Draw(gameTime);
        }
    
        /// <summary>
        /// This Event method will be called when the user changes the size of
        /// the window.
        /// </summary>
        /// <param name="sender">The caller of this method</param>
        /// <param name="e">The argument supplied by the caller.</param>
        void Window_ClientSizeChanged(object sender, EventArgs e)
        {
            m_userInput.updateCameraMatrices();       
        }
    }
}
