/************************************************************************
 * Copyright (c) 2008 by Gregor Ehrenstein (grx@pandacube.de)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ************************************************************************/

using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;

namespace BoxEdit3D
{
    /// <summary>
    /// The class that holds a list of the commands that where executed
    /// on the scene. It supports to undo and redo these commands. Commands 
    /// are issued for every action that changes the scene. This class is
    /// a support class for the command design pattern published by Gamma
    /// et. al.
    /// </summary>
    class CommandHistory
    {
        // list of all commands issued
        private ArrayList m_commandHistory = new ArrayList();

        // points to the current command (if undo was used
        // this pointer might not point to the end of the list)
        private int m_lastExecutedCommand = -1;

        /// <summary>
        /// Executes the given command and inserts the command
        /// into the command history. If undo was used before
        /// any commands that are "future" commands are deleted.
        /// </summary>
        /// <param name="command">The command that should be executed</param>
        public void executeAndAddCommand(Command command)
        {
            // is ArrayList empty?
            if (m_lastExecutedCommand != -1)
            {
                // is this command executed after some undo operations?
                if (m_lastExecutedCommand != (m_commandHistory.Count - 1))
                {
                    // remove all later commands
                    m_commandHistory.RemoveRange(m_lastExecutedCommand + 1, m_commandHistory.Count - 1 - m_lastExecutedCommand);
                }
            }

            // execute command
            command.execute();

            // add command to list and point the lastCommand index to the last element in the list
            m_commandHistory.Add(command);
            m_lastExecutedCommand = m_commandHistory.Count - 1;
        }

        /// <summary>
        /// Returns if undo operations are currently possible.
        /// </summary>
        /// <returns>True if a command execution can be undone.</returns>
        public bool isUndoPossible()
        {
            return (m_lastExecutedCommand >= 0);
        }

        /// <summary>
        /// Returns if redo operations are currently possible.
        /// </summary>
        /// <returns>True if a command can be redone.</returns>
        public bool isRedoPossible()
        {
            return (m_lastExecutedCommand < m_commandHistory.Count - 1);
        }

        /// <summary>
        /// Undo the latest command.
        /// </summary>
        public void undoCommand()
        {
            // check if undo is available
            if (isUndoPossible())
            {
                ((Command)m_commandHistory[m_lastExecutedCommand]).unexecute();
                m_lastExecutedCommand--;
            }
        }

        /// <summary>
        /// Redo the last undone command.
        /// </summary>
        public void redoCommand()
        {
            // check if redo is available
            if (isRedoPossible())
            {
                m_lastExecutedCommand++;
                ((Command)m_commandHistory[m_lastExecutedCommand]).execute();
            }
        }

        /// <summary>
        /// Clears the command history. Needs to be called after 
        /// the scene has been reset.
        /// </summary>
        public void ClearHistory()
        {
            m_commandHistory.Clear();
            m_lastExecutedCommand = -1;
        }
    }
}
