/************************************************************************
 * Copyright (c) 2008 by Gregor Ehrenstein (grx@pandacube.de)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ************************************************************************/

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;

namespace BoxEdit3D
{
    /// <summary>
    /// This class represents the floor of the scene. It renders
    /// a quad as floor. A special shader is used for the rendering 
    /// to let the floor appear to have a red grid on a translucent
    /// black surface.
    /// </summary>
    class Floor
    {
        // effect used to render the floor
        private Effect m_floorEffect;

        // size of the floor, effective size is double this value in x- 
        // and y-direction
        private const float m_size = 25.0f;

        // variables used to store the vertex information
        private VertexDeclaration m_vertexDeclaration;
        private VertexBuffer m_vertexBuffer;
        private VertexPositionTexture[] m_vertexData;

        // needed to render the floor
        private GraphicsDeviceManager m_graphics;

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="graphics">The GraphicsDeviceManager used to render the floor.</param>
        public Floor(GraphicsDeviceManager graphics)
        {
            m_graphics = graphics;
        }

        /// <summary>
        /// Defines the vertex information and loads the effect file
        /// used to render the floor. 
        /// </summary>
        /// <param name="content">The ContentManager used to load the effect file.</param>
        public void loadResources(ContentManager content)
        {
            // create 2 triangle spanning the floor
            m_vertexData = new VertexPositionTexture[4];
            m_vertexData[0].Position = new Vector3(-m_size, 0.0f, -m_size);
            m_vertexData[1].Position = new Vector3(m_size, 0.0f, -m_size);
            m_vertexData[2].Position = new Vector3(-m_size, 0.0f, m_size);
            m_vertexData[3].Position = new Vector3(m_size, 0.0f, m_size);
            m_vertexData[0].TextureCoordinate = new Vector2(0.0f, 1.0f);
            m_vertexData[1].TextureCoordinate = new Vector2(1.0f, 1.0f);
            m_vertexData[2].TextureCoordinate = new Vector2(0.0f, 0.0f);
            m_vertexData[3].TextureCoordinate = new Vector2(1.0f, 0.0f);

            // needed to render user vertices
            m_vertexDeclaration = new VertexDeclaration(m_graphics.GraphicsDevice, VertexPositionTexture.VertexElements);

            // create vertex buffer
            m_vertexBuffer = new VertexBuffer(m_graphics.GraphicsDevice, typeof(VertexPositionTexture), 4, BufferUsage.None);
            m_vertexBuffer.SetData<VertexPositionTexture>(m_vertexData);
            m_floorEffect = content.Load<Effect>("Shader\\FloorShader");
            m_floorEffect.Parameters["g_numberOfLines"].SetValue(m_size * 2.0f);
        }

        /// <summary>
        /// Renders the floor using the supplied view- and projection matrix.
        /// </summary>
        /// <param name="viewProjMatrix">The view and projection matrix.</param>
        public void render(Matrix viewProjMatrix)
        {
            // render floor

            // change render states and save the old values
            CullMode oldMode = m_graphics.GraphicsDevice.RenderState.CullMode;
            VertexDeclaration oldDecleration = m_graphics.GraphicsDevice.VertexDeclaration;
            m_graphics.GraphicsDevice.RenderState.CullMode = CullMode.None;
            bool alpha = m_graphics.GraphicsDevice.RenderState.AlphaBlendEnable;
            m_graphics.GraphicsDevice.RenderState.AlphaBlendEnable = true;
            m_graphics.GraphicsDevice.RenderState.SourceBlend = Blend.SourceAlpha;
            m_graphics.GraphicsDevice.RenderState.DestinationBlend = Blend.InverseSourceAlpha;

            // set correct vertex declaration
            m_graphics.GraphicsDevice.VertexDeclaration = m_vertexDeclaration;

            // set effect parameters
            m_floorEffect.Parameters["g_worldViewProj"].SetValue(viewProjMatrix);
            m_floorEffect.CommitChanges();

            // draw the quad
            m_floorEffect.Begin();
            m_floorEffect.Techniques[0].Passes[0].Begin();
            m_graphics.GraphicsDevice.DrawUserPrimitives<VertexPositionTexture>(PrimitiveType.TriangleStrip, m_vertexData, 0, 2);
            m_floorEffect.Techniques[0].Passes[0].End();
            m_floorEffect.End();

            // set old render states
            m_graphics.GraphicsDevice.RenderState.AlphaBlendEnable = alpha;
            m_graphics.GraphicsDevice.RenderState.CullMode = oldMode;
            m_graphics.GraphicsDevice.VertexDeclaration = oldDecleration;
        }
    }
}
