/************************************************************************
 * Copyright (c) 2008 by Gregor Ehrenstein (grx@pandacube.de)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ************************************************************************/

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Input;

namespace BoxEdit3D
{
    /// <summary>
    /// Derived class from the gizmo base class that can be used to
    /// rotate the box associated with this gizmo.
    /// </summary>
    class RotationGizmo : Gizmo
    {
        // static resources used to draw the gizmo
        private static Model m_rotationXModel;
        private static Model m_rotationYModel;
        private static Model m_rotationZModel;
        private static Effect m_rotationXEffect;
        private static Effect m_rotationYEffect;
        private static Effect m_rotationZEffect;

        // variables used to handle the dragging of the gizmo
        private bool m_dragging = false;
        private int m_dragDirection = -1;
        private Vector2 m_mouseStartPosition;
        private Quaternion m_oldRotation;
        private Quaternion m_newRotation;

        // factor used to scale the user input into object rotation
        private const float m_dragFactor = MathHelper.Pi / 1000.0f;

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="graphics">The GraphicsDeviceManager used to initialize the base class.</param>
        public RotationGizmo(GraphicsDeviceManager graphics)
            : base(graphics)
        {
        }

        /// <summary>
        /// Load the static resources (models and effects) that are used
        /// to draw the gizmo.
        /// </summary>
        /// <param name="content">The ContentManger used to load the resources.</param>
        public static void loadResources(ContentManager content)
        {
            // load static models
            m_rotationXModel = content.Load<Model>("Models\\rotationGizmoX");
            m_rotationYModel = content.Load<Model>("Models\\rotationGizmoY");
            m_rotationZModel = content.Load<Model>("Models\\rotationGizmoZ");
            m_rotationXEffect = content.Load<Effect>("Shader\\GizmoShaderX");
            m_rotationYEffect = content.Load<Effect>("Shader\\GizmoShaderY");
            m_rotationZEffect = content.Load<Effect>("Shader\\GizmoShaderZ");

            // set effects
            foreach (ModelMesh modmesh in m_rotationXModel.Meshes)
            {
                foreach (ModelMeshPart modmeshpart in modmesh.MeshParts)
                {
                    modmeshpart.Effect = m_rotationXEffect;
                }
            }
            foreach (ModelMesh modmesh in m_rotationYModel.Meshes)
            {
                foreach (ModelMeshPart modmeshpart in modmesh.MeshParts)
                {
                    modmeshpart.Effect = m_rotationYEffect;
                }
            }
            foreach (ModelMesh modmesh in m_rotationZModel.Meshes)
            {
                foreach (ModelMeshPart modmeshpart in modmesh.MeshParts)
                {
                    modmeshpart.Effect = m_rotationZEffect;
                }
            }
        }

        /// <summary>
        /// Renders the three models of the rotation gizmo.
        /// </summary>
        /// <param name="viewProj">The View and Projection Matrix used to render the gizmo.</param>
        public override void render(Matrix viewProj)
        {
            if (m_box != null)
            {
                // get model transform
                Matrix modelMatrix = m_box.getTransformMatrix();
                Matrix worldViewProj = modelMatrix * viewProj;

                // obtain the transform for each mesh
                Matrix[] transforms = new Matrix[m_rotationXModel.Bones.Count];
                m_rotationXModel.CopyAbsoluteBoneTransformsTo(transforms);

                // render each mesh
                foreach (ModelMesh mesh in m_rotationXModel.Meshes)
                {
                    foreach (Effect effect in mesh.Effects)
                    {
                        effect.Parameters["g_worldViewProj"].SetValue(transforms[mesh.ParentBone.Index] * worldViewProj);
                    }
                    mesh.Draw();
                }

                // obtain the transform for each mesh
                transforms = new Matrix[m_rotationYModel.Bones.Count];
                m_rotationYModel.CopyAbsoluteBoneTransformsTo(transforms);

                // render each mesh
                foreach (ModelMesh mesh in m_rotationYModel.Meshes)
                {
                    foreach (Effect effect in mesh.Effects)
                    {
                        effect.Parameters["g_worldViewProj"].SetValue(transforms[mesh.ParentBone.Index] * worldViewProj);
                    }
                    mesh.Draw();
                }

                // obtain the transform for each mesh
                transforms = new Matrix[m_rotationZModel.Bones.Count];
                m_rotationZModel.CopyAbsoluteBoneTransformsTo(transforms);

                // render each mesh
                foreach (ModelMesh mesh in m_rotationZModel.Meshes)
                {
                    foreach (Effect effect in mesh.Effects)
                    {
                        effect.Parameters["g_worldViewProj"].SetValue(transforms[mesh.ParentBone.Index] * worldViewProj);
                    }
                    mesh.Draw();
                }
            }
        }

        /// <summary>
        /// Undo any changes to the box done by the gizmo,
        /// as it has lost the focus.
        /// </summary>
        public override void lostFocus()
        {
            discardDrag();
            base.lostFocus();
        }

        /// <summary>
        /// Returns the number of gizmo parts that can be dragged.
        /// </summary>
        /// <returns>The number of parts of the gizmo that can be dragged.</returns>
        public override int getNumberOfGizmoParts()
        {
            return 3;
        }

        /// <summary>
        /// Returns the model of the given gizmo part. This model can then be used to
        /// check if it was picked.
        /// </summary>
        /// <param name="index">The index of the gizmo part.</param>
        /// <returns>The model of the given gizmo part.</returns>
        public override Model getGizmoPart(int index)
        {
            switch (index)
            {
                case 0:
                    return m_rotationXModel;
                case 1:
                    return m_rotationYModel;
                case 2:
                    return m_rotationZModel;
                default:
                    return null;
            }
        }

        /// <summary>
        /// Start dragging on the gizmo. The given gizmoPart number indentifies which
        /// part of the gizmo is dragged.
        /// </summary>
        /// <param name="gizmoPart">The number of the gizmo part that is dragged.</param>
        public override void startDrag(int gizmoPart)
        {
            if (!m_dragging)
            {
                if (!((gizmoPart >= 0) && (gizmoPart <= 2))) return;
                m_dragDirection = gizmoPart;

                // save old rotation to undo changes if drag is aborted
                m_oldRotation = m_box.getRotation();
                m_newRotation = m_oldRotation;
                MouseState mouseState = Mouse.GetState();
                m_mouseStartPosition = new Vector2(mouseState.X, mouseState.Y);
                m_dragging = true;
            }
        }

        /// <summary>
        /// Apply the the changed done by the gizmo and issue a command that
        /// capsulates the changes.
        /// </summary>
        /// <returns>The command that capsulates the changes to the box.</returns>
        public override Command applyDrag()
        {
            if (m_dragging)
            {
                m_dragDirection = -1;
                // set old rotation, as the command saves the previous state
                m_box.setRotation(m_oldRotation);

                // issue a rotate command
                RotateCommand command = new RotateCommand(m_box, m_newRotation);

                // reset variables
                m_oldRotation = Quaternion.Identity;
                m_newRotation = Quaternion.Identity;
                m_mouseStartPosition = Vector2.Zero;
                m_dragging = false;

                return command;
            }
            else return null;
        }

        /// <summary>
        /// Discard the changes done by the user and reset the values of the
        /// used box.
        /// </summary>
        public override void discardDrag()
        {
            if (m_dragging)
            {
                // reset variables
                m_dragDirection = -1;
                m_box.setRotation(m_oldRotation);
                m_oldRotation = Quaternion.Identity;
                m_newRotation = Quaternion.Identity;
                m_mouseStartPosition = Vector2.Zero;
                m_dragging = false;
            }
        }

        /// <summary>
        /// Rotate the box depending on the mouse moves of the user when in drag mode.
        /// </summary>
        public override void handleMouseInput(Matrix viewProjectionMatrix)
        {
            if (m_dragging)
            {
                // calculate mouse movement
                MouseState mouseState = Mouse.GetState();
                Vector2 delta = new Vector2(mouseState.X, mouseState.Y) - m_mouseStartPosition;
                m_newRotation = m_oldRotation;
                Quaternion newRotation = Quaternion.Identity;

                float deltaUsed = delta.X - delta.Y;

                // rotate around an axis
                switch (m_dragDirection)
                {
                    case 0:
                        newRotation = Quaternion.CreateFromAxisAngle(new Vector3(1.0f, 0.0f, 0.0f), deltaUsed * m_dragFactor);
                        break;
                    case 1:
                        newRotation = Quaternion.CreateFromAxisAngle(new Vector3(0.0f, 1.0f, 0.0f), deltaUsed * m_dragFactor);
                        break;
                    case 2:
                        newRotation = Quaternion.CreateFromAxisAngle(new Vector3(0.0f, 0.0f, 1.0f), deltaUsed * m_dragFactor);
                        break;
                }

                // apply rotation
                m_newRotation *= newRotation;
                m_newRotation.Normalize();

                m_box.setRotation(m_newRotation);
            }
        }
    }
}
