/************************************************************************
 * Copyright (c) 2008 by Gregor Ehrenstein (grx@pandacube.de)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ************************************************************************/

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Input;

namespace BoxEdit3D
{
    /// <summary>
    /// Derived class from the gizmo base class that can be used to
    /// scale the box associated with this gizmo.
    /// </summary>
    class ScaleGizmo : Gizmo
    {
        // static resources used to draw the gizmo
        private static Model m_scaleXModel;
        private static Model m_scaleYModel;
        private static Model m_scaleZModel;
        private static Effect m_scaleXEffect;
        private static Effect m_scaleYEffect;
        private static Effect m_scaleZEffect;

        // variables used to handle the dragging of the gizmo
        private bool m_dragging = false;
        private int m_dragDirection = -1;
        private Vector2 m_mouseStartPosition;
        private Vector3 m_oldScale;
        private Vector3 m_newScale;

        // factor used to scale the user input into object scale
        private const float m_dragFactor = 0.01f;

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="graphics">The GraphicsDeviceManager used to initialize the base class.</param>
        public ScaleGizmo(GraphicsDeviceManager graphics)
            : base(graphics)
        {
        }

        /// <summary>
        /// Load the static resources (models and effects) that are used
        /// to draw the gizmo.
        /// </summary>
        /// <param name="content">The ContentManger used to load the resources.</param>
        public static void loadResources(ContentManager content)
        {
            // load static models
            m_scaleXModel = content.Load<Model>("Models\\scaleGizmoX");
            m_scaleYModel = content.Load<Model>("Models\\scaleGizmoY");
            m_scaleZModel = content.Load<Model>("Models\\scaleGizmoZ");
            m_scaleXEffect = content.Load<Effect>("Shader\\GizmoShaderX");
            m_scaleYEffect = content.Load<Effect>("Shader\\GizmoShaderY");
            m_scaleZEffect = content.Load<Effect>("Shader\\GizmoShaderZ");

            // set effect to each mesh
            foreach (ModelMesh modmesh in m_scaleXModel.Meshes)
            {
                foreach (ModelMeshPart modmeshpart in modmesh.MeshParts)
                {
                    modmeshpart.Effect = m_scaleXEffect;
                }
            }
            foreach (ModelMesh modmesh in m_scaleYModel.Meshes)
            {
                foreach (ModelMeshPart modmeshpart in modmesh.MeshParts)
                {
                    modmeshpart.Effect = m_scaleYEffect;
                }
            }
            foreach (ModelMesh modmesh in m_scaleZModel.Meshes)
            {
                foreach (ModelMeshPart modmeshpart in modmesh.MeshParts)
                {
                    modmeshpart.Effect = m_scaleZEffect;
                }
            }
        }

        /// <summary>
        /// Renders the three models of the scale gizmo.
        /// </summary>
        /// <param name="viewProj">The View and Projection Matrix used to render the gizmo.</param>
        public override void render(Matrix viewProj)
        {
            if (m_box != null)
            {
                Matrix modelMatrix = m_box.getTransformMatrix();
                Matrix worldViewProj = modelMatrix * viewProj;

                // obtain the transform for each mesh
                Matrix[] transforms = new Matrix[m_scaleXModel.Bones.Count];
                m_scaleXModel.CopyAbsoluteBoneTransformsTo(transforms);

                // render each mesh
                foreach (ModelMesh mesh in m_scaleXModel.Meshes)
                {
                    foreach (Effect effect in mesh.Effects)
                    {
                        effect.Parameters["g_worldViewProj"].SetValue(transforms[mesh.ParentBone.Index] * worldViewProj);
                    }
                    mesh.Draw();
                }

                // obtain the transform for each mesh
                transforms = new Matrix[m_scaleYModel.Bones.Count];
                m_scaleYModel.CopyAbsoluteBoneTransformsTo(transforms);

                // render each mesh
                foreach (ModelMesh mesh in m_scaleYModel.Meshes)
                {
                    foreach (Effect effect in mesh.Effects)
                    {
                        effect.Parameters["g_worldViewProj"].SetValue(transforms[mesh.ParentBone.Index] * worldViewProj);
                    }
                    mesh.Draw();
                }

                // obtain the transform for each mesh
                transforms = new Matrix[m_scaleZModel.Bones.Count];
                m_scaleZModel.CopyAbsoluteBoneTransformsTo(transforms);

                // render each mesh
                foreach (ModelMesh mesh in m_scaleZModel.Meshes)
                {
                    foreach (Effect effect in mesh.Effects)
                    {
                        effect.Parameters["g_worldViewProj"].SetValue(transforms[mesh.ParentBone.Index] * worldViewProj);
                    }
                    mesh.Draw();
                }
            }
        }

        /// <summary>
        /// Undo any changes to the box done by the gizmo,
        /// as it has lost the focus.
        /// </summary>
        public override void lostFocus()
        {
            discardDrag();
            base.lostFocus();
        }

        /// <summary>
        /// Returns the number of gizmo parts that can be dragged.
        /// </summary>
        /// <returns>The number of parts of the gizmo that can be dragged.</returns>
        public override int getNumberOfGizmoParts()
        {
            return 3;
        }

        /// <summary>
        /// Returns the model of the given gizmo part. This model can then be used to
        /// check if it was picked.
        /// </summary>
        /// <param name="index">The index of the gizmo part.</param>
        /// <returns>The model of the given gizmo part.</returns>
        public override Model getGizmoPart(int index)
        {
            switch (index)
            {
                case 0:
                    return m_scaleXModel;
                case 1:
                    return m_scaleYModel;
                case 2:
                    return m_scaleZModel;
                default:
                    return null;
            }
        }

        /// <summary>
        /// Start dragging on the gizmo. The given gizmoPart number indentifies which
        /// part of the gizmo is dragged.
        /// </summary>
        /// <param name="gizmoPart">The number of the gizmo part that is dragged.</param>
        public override void startDrag(int gizmoPart)
        {
            if (!m_dragging)
            {
                if (!((gizmoPart >= 0) && (gizmoPart <= 2))) return;
                m_dragDirection = gizmoPart;

                // save old scale so the changes can be undone
                m_oldScale = m_box.getScale();
                m_newScale = m_oldScale;
                MouseState mouseState = Mouse.GetState();
                m_mouseStartPosition = new Vector2(mouseState.X, mouseState.Y);
                m_dragging = true;
            }
        }

        /// <summary>
        /// Apply the the changed done by the gizmo and issue a command that
        /// capsulates the changes.
        /// </summary>
        /// <returns>The command that capsulates the changes to the box.</returns>
        public override Command applyDrag()
        {
            if (m_dragging)
            {
                m_dragDirection = -1;
                // apply old scale as the command saves the previous state
                m_box.setScale(m_oldScale);

                // issue scale command
                ScaleCommand command = new ScaleCommand(m_box, m_newScale);

                // reset variables
                m_oldScale = Vector3.One;
                m_newScale = Vector3.One;
                m_mouseStartPosition = Vector2.Zero;
                m_dragging = false;

                return command;
            }
            else return null;
        }

        /// <summary>
        /// Discard the changes done by the user and reset the values of the
        /// used box.
        /// </summary>
        public override void discardDrag()
        {
            if (m_dragging)
            {
                // reset variables
                m_dragDirection = -1;
                m_box.setScale(m_oldScale);
                m_oldScale = Vector3.One;
                m_newScale = Vector3.One;
                m_mouseStartPosition = Vector2.Zero;
                m_dragging = false;
            }
        }

        /// <summary>
        /// Scale the box depending on the mouse moves of the user when in drag mode.
        /// </summary>
        public override void handleMouseInput(Matrix viewProjectionMatrix)
        {
            if (m_dragging)
            {
                // calculate mouse movement
                MouseState mouseState = Mouse.GetState();
                Vector2 delta = new Vector2(mouseState.X, mouseState.Y) - m_mouseStartPosition;
                delta.Y = -delta.Y; 
                m_newScale = m_oldScale;

                // scale in one direction
                switch (m_dragDirection)
                {
                    case 0:
                        Vector4 dir1 = new Vector4(0.0f, 0.0f, 0.0f, 1.0f);
                        Vector4 dir2 = new Vector4(1.0f, 0.0f, 0.0f, 1.0f);
                        dir1 = Vector4.Transform(dir1, getTransformMatrix());
                        dir2 = Vector4.Transform(dir2, getTransformMatrix());
                        dir1 = Vector4.Transform(dir1, viewProjectionMatrix);
                        dir2 = Vector4.Transform(dir2, viewProjectionMatrix);
                        dir1 /= dir1.W;
                        dir2 /= dir2.W;
                        Vector2 dir_2 = new Vector2(dir2.X - dir1.X, dir2.Y - dir1.Y);
                        if (dir_2.Length() == 0.0f) dir_2 = new Vector2(1.0f, 0.0f);
                        dir_2.Normalize();
                        float deltaUsed = Vector2.Dot(delta, dir_2);
                        m_newScale.X += deltaUsed * m_dragFactor;
                        break;
                    case 1:
                        dir1 = new Vector4(0.0f, 0.0f, 0.0f, 1.0f);
                        dir2 = new Vector4(0.0f, 1.0f, 0.0f, 1.0f);
                        dir1 = Vector4.Transform(dir1, getTransformMatrix());
                        dir2 = Vector4.Transform(dir2, getTransformMatrix());
                        dir1 = Vector4.Transform(dir1, viewProjectionMatrix);
                        dir2 = Vector4.Transform(dir2, viewProjectionMatrix);
                        dir1 /= dir1.W;
                        dir2 /= dir2.W;
                        dir_2 = new Vector2(dir2.X - dir1.X, dir2.Y - dir1.Y);
                        if (dir_2.Length() == 0.0f) dir_2 = new Vector2(1.0f, 0.0f);
                        dir_2.Normalize();
                        deltaUsed = Vector2.Dot(delta, dir_2);
                        m_newScale.Y += deltaUsed * m_dragFactor;
                        break;
                    case 2:
                        dir1 = new Vector4(0.0f, 0.0f, 0.0f, 1.0f);
                        dir2 = new Vector4(0.0f, 0.0f, 1.0f, 1.0f);
                        dir1 = Vector4.Transform(dir1, getTransformMatrix());
                        dir2 = Vector4.Transform(dir2, getTransformMatrix());
                        dir1 = Vector4.Transform(dir1, viewProjectionMatrix);
                        dir2 = Vector4.Transform(dir2, viewProjectionMatrix);
                        dir1 /= dir1.W;
                        dir2 /= dir2.W;
                        dir_2 = new Vector2(dir2.X - dir1.X, dir2.Y - dir1.Y);
                        if (dir_2.Length() == 0.0f) dir_2 = new Vector2(1.0f, 0.0f);
                        dir_2.Normalize();
                        deltaUsed = Vector2.Dot(delta, dir_2);
                        m_newScale.Z += deltaUsed * m_dragFactor;
                        break;
                }

                // check if new scale is valid
                if (m_newScale.X <= 0.001f) m_newScale.X = 0.001f;
                if (m_newScale.Y <= 0.001f) m_newScale.Y = 0.001f;
                if (m_newScale.Z <= 0.001f) m_newScale.Z = 0.001f;              

                m_box.setScale(m_newScale);
            }
        }
    }
}
