/************************************************************************
 * Copyright (c) 2008 by Gregor Ehrenstein (grx@pandacube.de)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ************************************************************************/

using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using System.Xml.Serialization;
using System.IO;
using Microsoft.Xna.Framework;

namespace BoxEdit3D
{
    /// <summary>
    /// This class capsulates the configuration of the scene. It holds
    /// all current boxes added to the scene.
    /// </summary>
    class Scene
    {
        // used to keep track of the boxes
        private ArrayList m_boxes = new ArrayList();
        private ArrayList m_xmlBoxes;
        private int m_selectedBox = -1;

        /// <summary>
        /// Adds the supplied box to the scene.
        /// </summary>
        /// <param name="box">The box that should be added.</param>
        public void addBox(Box box)
        {
            // fresh box are selected
            box.setSelectedStatus(true);

            // add box to list
            m_boxes.Add(box);

            // deselect previously selected box
            if ((m_selectedBox >= 0) && (m_selectedBox < m_boxes.Count)) ((Box)m_boxes[m_selectedBox]).setSelectedStatus(false);

            // save which box is selected
            m_selectedBox = m_boxes.Count - 1;
        }

        /// <summary>
        /// Remove the supplied box from the scene.
        /// </summary>
        /// <param name="box">The box that should be removed from the scene.</param>
        public void removeBox(Box box)
        {
            // deselect box
            if ((m_selectedBox >= 0) && (m_selectedBox < m_boxes.Count)) ((Box)m_boxes[m_selectedBox]).setSelectedStatus(false);

            // remove box
            m_boxes.Remove(box);
            m_selectedBox = -1;
        }

        /// <summary>
        /// Select the box with the given index.
        /// </summary>
        /// <param name="index">The index of the box that should be selected.</param>
        public void selectBox(int index)
        {
            // deselect previously selected box
            if ((m_selectedBox >= 0) && (m_selectedBox < m_boxes.Count))
            {
                ((Box)m_boxes[m_selectedBox]).setSelectedStatus(false);
                m_selectedBox = -1;
            }

            // check bounds of index
            if ((index >= 0) && (index < m_boxes.Count))
            {
                // select new box
                m_selectedBox = index;
                ((Box)m_boxes[m_selectedBox]).setSelectedStatus(true);
            }
        }

        /// <summary>
        /// Returns the currently selected box.
        /// </summary>
        /// <returns>Returns the currently selected box. If no box is selected the value null is returned.</returns>
        public Box getSelectedBox()
        {
            if ((m_selectedBox >= 0) && (m_selectedBox < m_boxes.Count)) return ((Box)m_boxes[m_selectedBox]);
            else return null;
        }

        /// <summary>
        /// Returns the index of the currently selected box. If no box is selected
        /// the value -1 is returned.
        /// </summary>
        /// <returns>The index of the currently selected box.</returns>
        public int getSelectedBoxNumber()
        {
            return m_selectedBox;
        }

        /// <summary>
        /// Returns the number of boxes that are currently in the scene.
        /// </summary>
        /// <returns>The number of boxes displayed in the scene.</returns>
        public int numberOfBoxes()
        {
            return m_boxes.Count;
        }

        /// <summary>
        /// Returns if a box is selected.
        /// </summary>
        /// <returns>True if a box is selected.</returns>
        public bool isBoxSelected()
        {
            return (m_selectedBox != -1);
        }

        /// <summary>
        /// Returns the box indicated by the supplied index.
        /// </summary>
        /// <param name="index">The index of the box that should be supplied</param>
        /// <returns>The indicated box. If the index is out of the range of the currently 
        /// held boxes in the scene the value null is returned.</returns>
        public Box getBox(int index)
        {
            if ((index >= 0) && (index < m_boxes.Count)) return ((Box)m_boxes[index]);
            else return null;
        }

        /// <summary>
        /// Save the current scene to an xml file indicated by the supplied file name.
        /// </summary>
        /// <param name="filename">The name of the xml-file.</param>
        public void saveSceneToXML(string filename)
        {
            // create list of xmlBoxes
            m_xmlBoxes = new ArrayList(m_boxes.Count);
            foreach(Box box in m_boxes) {
                XMLBox xmlBox = box.getXMLBox();
                m_xmlBoxes.Add(xmlBox);
            }

            // serialize into xml file
            Type[] typeList = new Type[1];
            typeList[0] = typeof(XMLBox);
            XmlSerializer serializer = new XmlSerializer(typeof(ArrayList), typeList);
            TextWriter writer = new StreamWriter(filename);
            serializer.Serialize(writer, m_xmlBoxes);
            writer.Close();
        }

        /// <summary>
        /// Loads the scene from the supplied xml-file.
        /// </summary>
        /// <param name="graphics">The GraphicsDeviceManager needed to create the box instances.</param>
        /// <param name="filename">The name of the xml-file.</param>
        /// <returns>True if it was possible to read the scene from the xml file.</returns>
        public bool loadSceneFromXML(GraphicsDeviceManager graphics, string filename)
        {
            // deserialize from xml file
            Type[] typeList = new Type[1];
            typeList[0] = typeof(XMLBox);
            XmlSerializer serializer = new XmlSerializer(typeof(ArrayList), typeList);
            serializer.UnknownNode += new XmlNodeEventHandler(serializer_UnknownNode);
            serializer.UnknownAttribute += new XmlAttributeEventHandler(serializer_UnknownAttribute);
            
            // open file
            XmlReaderSettings settings = new XmlReaderSettings();
            settings.ConformanceLevel = ConformanceLevel.Fragment;
            settings.IgnoreWhitespace = true;
            settings.IgnoreComments = true;
            XmlReader reader = XmlReader.Create(filename, settings);

            // check if file has correct format
            if (!serializer.CanDeserialize(reader))
            {
                reader.Close();
                return false;
            }

            // deserialize
            m_xmlBoxes = (ArrayList) serializer.Deserialize(reader);
            reader.Close();
            
            // create list of boxes
            m_boxes = new ArrayList(m_xmlBoxes.Count);
            foreach (XMLBox xmlBox in m_xmlBoxes)
            {
                Box box = new Box(graphics, xmlBox);
                m_boxes.Add(box);
            }

            // no box is selected after load
            m_selectedBox = -1;

            return true;            
        }
 
        /// <summary>
        /// This event method will be called by the xml-deserializer if unknow nodes are found.
        /// </summary>
        /// <param name="sender">The sender that is calling this method.</param>
        /// <param name="e">Event argument that was supplied by the caller.</param>
        private void serializer_UnknownNode(object sender, XmlNodeEventArgs e)
        {
            Console.WriteLine("Unknown Node:" +   e.Name + "\t" + e.Text);
        }

        /// <summary>
        /// This event method will be called by the xml-deserializer if unknow attributes are found.
        /// </summary>
        /// <param name="sender">The sender that is calling this method.</param>
        /// <param name="e">Event argument that was supplied by the caller.</param>
        private void serializer_UnknownAttribute(object sender, XmlAttributeEventArgs e)
        {
            System.Xml.XmlAttribute attr = e.Attr;
            Console.WriteLine("Unknown attribute " + attr.Name + "='" + attr.Value + "'");
        }

        /// <summary>
        /// Renders the current scene using the supplied matrix.
        /// </summary>
        /// <param name="viewProjectionMatrix">The view- and projection matrix that is used to render the scene.</param>
        public void render(Matrix viewProjectionMatrix)
        {
            foreach (Box box in m_boxes)
            {
                box.render(viewProjectionMatrix);
            }
        }

        /// <summary>
        /// Resets the current scene. This will remove all boxes in the scene.
        /// </summary>
        public void reset()
        {
            m_boxes.Clear();
            m_selectedBox = -1;
        }
    }
}
