/************************************************************************
 * Copyright (c) 2008 by Gregor Ehrenstein (grx@pandacube.de)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ************************************************************************/

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace BoxEdit3D
{
    /// <summary>
    /// This class can be used to ensure that each texture is only created
    /// once. The TextureManager stores the textures internally. If a texture
    /// is requested, that is not loaded yet, it will be loaded from disc and 
    /// added to the internal data structure. If a certain texture can not be
    /// found or an error occurs during load, then the TextureManager will
    /// return a null reference.
    /// </summary>
    class TextureManager
    {
        // stores the textures in a hash map
        private Dictionary<string, Texture2D> m_textures = new Dictionary<string,Texture2D>();

        // needed to create the textures
        private GraphicsDeviceManager m_graphics;

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="graphics">The GraphicsDeviceManager is used to create the textures.</param>
        public TextureManager(GraphicsDeviceManager graphics)
        {
            m_graphics = graphics;
        }
        
        /// <summary>
        /// Returns the texture indicated by the supplied filename. If the texture
        /// was already loaded it returns this texture. Otherwise the texture will
        /// be loaded from file and returned to the caller.
        /// </summary>
        /// <param name="filename">The filename of the requested texture.</param>
        /// <returns>The requested texture object. If the file could not be loaded this value is null.</returns>
        public Texture2D getTexture(string filename) 
        {
            Texture2D result = null;
            // check if the filename is not empty
            if (filename != "") {
                // try to find the texture in the hash map
                if (!m_textures.TryGetValue(filename, out result)) {
                    // try to load the texture and add it to the dictonary
                    if (File.Exists(filename)){
                        try {
                            result = Texture2D.FromFile(m_graphics.GraphicsDevice, filename);
                        }
                        catch(Exception e) {
                            Console.WriteLine("Exception caught while loading a texture: " + e.Message);
                            result = null;
                        }
                        if (result != null) {
                            // add texture to hash map
                            m_textures.Add(filename, result);
                        }
                    }

                }
            }
            return result;
        }

        /// <summary>
        /// This method frees the resources taken up by the textures in the internal data structure.
        /// </summary>
        public void freeResources()
        {
            m_textures.Clear();
        }
    }
}
